/*
Copyright (C) 2004-2005 Amour Hassan

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

#ifndef _RG3DROM_H
#define _RG3DROM_H

#include <vector>
#include <rgPort.h>
#include <rgStandard.h>
#include <rgOpenGL.h>

#define SK3DROM_SKEL_FRAME_NAME_LEN	72
#define SK3DROM_SEQUENCE_NAME_LEN	72

class rgBSP_NTTC_ExtData;

template <class T>
class rgBSPTree;

class rgCSGBrush;
class rgCSGWorld;

class sk3DRomModel;
class sk3DRomTag;
class sk3DRomSkel;

class sk3DRomDrawContext;

#define SK3DROM_VERT_HIDDEN_BIT		0x0001
#define SK3DROM_VERT_SELECTED_BIT	0x0002
#define SK3DROM_VERTEX_VERSION_A	0x1
class sk3DRomVertex : public rgVector
{
public:
	float weight;

	friend class sk3DRomModel;


	sk3DRomVertex();
	sk3DRomVertex( const float p_x, const float p_y, const float p_z);
	sk3DRomVertex( const rgVector &v);
	sk3DRomVertex &operator =( const rgVector &v);

	int gid() const { return m_gid;}
	int bid() const { return m_bid;}

	void hide() { flags |= SK3DROM_VERT_HIDDEN_BIT;}
	void reveal() { flags &= ~SK3DROM_VERT_HIDDEN_BIT;}
	int hidden() const { return (flags & SK3DROM_VERT_HIDDEN_BIT);}
	int selected() const { return (flags & SK3DROM_VERT_SELECTED_BIT);}

protected:
	// Flags:
	//
	int m_bid;	// Bone ID
	int m_gid;	// Group ID
	int flags;
};

#define SK3DROM_GROUP_HIDDEN_BIT	0x01
#define SK3DROM_GROUP_NAME_MAX_LEN	80
class sk3DRomGroup
{
public:
	char gname[ SK3DROM_GROUP_NAME_MAX_LEN];
	unsigned char gcolor[4];	// RGBA

	// Flags:
	// Bit 0: 0 = Group is visible, 1 = group is hidden
	int flags;
	std::vector <int> vid;	// List of vertices belonging to the group

	sk3DRomGroup();
	int hidden() const { return (flags & SK3DROM_GROUP_HIDDEN_BIT);};

	void set_color3() const {
		glColor3ub( gcolor[0], gcolor[1], gcolor[2]);}
	void set_color4() const {
		glColor4ub( gcolor[0], gcolor[1], gcolor[2], gcolor[3]);}
};

#define SK3DROM_FACE_FLAT_BIT	0x01	// Face must be flat shaded
#define SK3DROM_FACE_ALPHA_BIT	0x02	// Face makes use of an alpha component
#define SK3DROM_FACE_BSP_BIT	0x04	// Face is a member of a BSP tree. This
					// means that it should only be
					// rendered by the BSP tree, rather
					// than the standard rendering
					// functions.
#define SK3DROM_FACE_COLOR_BIT	0x08	// Face uses its color component. If
					// this bit is not set, then alpha
					// color component has no effect.
#define SK3DROM_FACE_MAPPED_BIT	0x10	// The face belongs to at least one
					// texture page.
#define SK3DROM_FACE_VERSION_A	0
#define SK3DROM_FACE_VERSION_B	1
class sk3DRomFace
{
public:
	short vid[3];		// Vertex index values
	rgVector n[3];		// Vertex normals
	unsigned char li[12];	// Per vertex lighting ((R,G,B,A) per vertex)
	short tid;		// Single texture index value
	float map[6];		// Mapping coordinates ((U,V) per vertex).
				// NOTE: Since it is desirable to have more
				//	than one texture per face, additional
				//	mapping coordinates are stored else
				//	where. This allows for an unlimited
				//	number of mapping coordinates. Infact,
				//	all mapping coordinates should be
				//	stored elsewhere, but single texture
				//	mapping coordinates are stored here
				//	for backwards compatibility.
	int flags;
	int gid;

	sk3DRomFace();

	void set_bsp() { flags |= SK3DROM_FACE_BSP_BIT;}
	void unset_bsp() { flags &= ~SK3DROM_FACE_BSP_BIT;}
	void set_use_color() { flags |= SK3DROM_FACE_COLOR_BIT;}
	void unset_use_color() { flags &= ~SK3DROM_FACE_COLOR_BIT;}
	int use_color() const { return (flags & SK3DROM_FACE_COLOR_BIT);}
	int is_alpha() const { return (flags & SK3DROM_FACE_ALPHA_BIT);}
	int is_bsp() const { return (flags & SK3DROM_FACE_BSP_BIT);}
	void force_flat_shade() { flags |= SK3DROM_FACE_FLAT_BIT;}
	void unforce_flat_shade() {flags &= ~SK3DROM_FACE_FLAT_BIT;}
	bool mapped() const {
		if( (flags & SK3DROM_FACE_MAPPED_BIT)) return true;
		return false;
	}
	void flip() { rgSwap( vid[0], vid[2]);};
	//void gen_world_map();
	void gen_planar_map( const sk3DRomModel &mdl,
        const rgCoordSystem &cs);
};

#define SK3DROM_TEXTURE_SMOOTH_BIT	0x00000001
#define SK3DROM_TEXTURE_MIPMAP_BIT	0x00000002
#define SK3DROM_TEXTURE_COLORKEY_BIT	0x00000004
#define SK3DROM_TEXTURE_ALPHA_BIT	0x00000008	// Face makes use of an
							// alpha component.

#define SK3DROM_TEXTURE_VERSION_A	0x0
#define SK3DROM_TEXTURE_VERSION_B	0x1
class sk3DRomTexture
{
public:
	friend class sk3DRomModel;

	std::string m_tname;	// The name of the texture

	sk3DRomTexture();

	int load();
	int load( const char *p_fname);
	void clear();

	void filename( char *des, const int len) const {
		rg_get_unix_file_name_only( des, m_fname.c_str(), len);
	}

	void sk3DRomTexture::pathname( char *des, const int len) const {
		rg_get_unix_path_name_only( des, m_fname.c_str(), len);
	}

	const char *full_path() const { return m_fname.c_str();}
	const bool is_smooth() const {
		if( (m_flags & SK3DROM_TEXTURE_SMOOTH_BIT)) return true;
		return false;
	}
	const bool is_mipmapped() const {
		if( (m_flags & SK3DROM_TEXTURE_MIPMAP_BIT)) return true;
		return false;
	}
	const bool is_alpha() const {
		if( (m_flags & SK3DROM_TEXTURE_ALPHA_BIT)) return true;
		return false;
	}

	void set_mipmap( const bool b) {
		if( b) m_flags |= SK3DROM_TEXTURE_MIPMAP_BIT;
		else m_flags &= ~SK3DROM_TEXTURE_MIPMAP_BIT;
	}

	void set_smooth( const bool b) {
		if( b) m_flags |= SK3DROM_TEXTURE_SMOOTH_BIT;
		else m_flags &= ~SK3DROM_TEXTURE_SMOOTH_BIT;
	}

	void set_alpha( const bool b) {
		if( b) m_flags |= SK3DROM_TEXTURE_ALPHA_BIT;
		else m_flags &= ~SK3DROM_TEXTURE_ALPHA_BIT;
	}

	int full_path_length() const { return int( m_fname.size());}
	int flags() const { return m_flags;}

	// TODO: These should be protected.
	void set_file_name( const char *p_fname) { m_fname = p_fname;}
	void set_flags( const int f) { m_flags = f;}

	void bind() const {
		if( m_tex_alternate) m_tex_alternate->bind();
		else m_tex.bind();
	}

	rgGLTexture *alternate_texture() { return m_tex_alternate;}

	int texture_id() const { return m_tex.getTextureID();}
	int width() const { return m_tex.getWidth();}
	int height() const { return m_tex.getHeight();}

	void set_alternate_texture( rgGLTexture *t) {
		m_tex_alternate = t;}

	unsigned char m_colorkey[4];

protected:
	// File input\output functions.
	int read_file_name( std::istream &inf, const int len);
	int read_tex_name( std::istream &inf, const int len);

private:
	std::string m_fname;	// The file name of the texture
	int m_flags;		// Texture attribute flags
	rgGLTexture m_tex;
	rgGLTexture *m_tex_alternate;
};

class SelectionCache
{
public:
	SelectionCache() { start_selection();}
	void clear() { sel_list.clear();}
	void start_selection() { pre = -1;}
	void add_selection( const int i) { sel_list.push_back( i);}
	int roll_over_selection() {
		if( (pre >= 0) && !(sel_list.empty())) {
			const int tmp = sel_list[0];
			sel_list.clear();
			sel_list.push_back( tmp);
			return tmp;
		}
		return -1;
	}
	bool was_preselected( const int i) {
		if( pre < 0) pre = 0;			// First selection
		if( pre >= int( sel_list.size())) {	// New overlap
			sel_list.push_back( i);
			return false;
		}
		if( i == sel_list[ pre]) {		// Preselected
			pre++;
			return true;
		}
		sel_list.clear();			// New fresh selection
		sel_list.push_back( i);
		pre = 0;
		return false;
	}

private:
	std::vector <int> sel_list;
	int pre;
};

class sk3DRomTexturePageSelVertex
{
public:
	int fid;
	int vid;

	sk3DRomTexturePageSelVertex() { fid = vid = -1;}

	sk3DRomTexturePageSelVertex( const int p_fid, const int p_vid) {
		fid = p_fid;
		vid = p_vid;
	}
};

class sk3DRomTextureFace
{
public:

	int fid() const { return m_fid;}

	const rgVector &v( const int i) const {
		if( i < 0) return m_v[0];
		else if( i >= 3) return m_v[2];
		return m_v[i];
	}

	void set_fid( const int fid) { m_fid = fid;}

	sk3DRomTextureFace() { m_fid = -1;}
	sk3DRomTextureFace( const int fid, const rgVector a, const rgVector b,
		const rgVector c);

	void transform_vertex( const int vid, const rgMatrix &m) {
		if( (vid < 0) || (vid >= 3)) return;
		m_v[ vid] = m.pointMul( m_v[ vid]);
	}

	void update_vertex( const int vid, const rgVector p) {
		if( (vid < 0) || (vid >= 3)) return;
		m_v[ vid] = p;	
	}

	void update_vertices( const rgVector a, const rgVector b,
		const rgVector c) {
		m_v[0] = a;
		m_v[1] = b;
		m_v[2] = c;
	}


private:
	int m_fid;
	rgVector m_v[3];
};

#define SK3DROM_TEX_PAGE_NAME_LEN 80
#define SK3DROM_PAGE_VERSION_A	1

#define SK3DROM_PAGE_WORLD_MAP_BIT	0x1
#define SK3DROM_PAGE_LOCAL_MAP_BIT	0x2
#define SK3DROM_PAGE_WORLD_TEXTURE_BIT	0x4
#define SK3DROM_PAGE_ENABLED_BIT	0x8
class sk3DRomTexturePage
{
public:

	sk3DRomTexturePage();

	int world_texture() const {
		return (m_flags & SK3DROM_PAGE_WORLD_TEXTURE_BIT);}
	int default_map() const { if( world_map() || local_map()) return false;
		return true;}
	int world_map() const { return (m_flags & SK3DROM_PAGE_WORLD_MAP_BIT);}
	int local_map() const { return (m_flags & SK3DROM_PAGE_LOCAL_MAP_BIT);}

	void set_world_texture() {
		m_flags |= SK3DROM_PAGE_WORLD_TEXTURE_BIT;}
	void unset_world_texture() {
		m_flags &= ~SK3DROM_PAGE_WORLD_TEXTURE_BIT;}

	void set_default_map() {
		m_flags &= ~(SK3DROM_PAGE_WORLD_MAP_BIT |
			SK3DROM_PAGE_LOCAL_MAP_BIT);
	}
	void set_world_map() {
		set_default_map();
		m_flags |= SK3DROM_PAGE_WORLD_MAP_BIT;
	}
	void set_local_map() {
		set_default_map();
		m_flags |= SK3DROM_PAGE_LOCAL_MAP_BIT;
	}

	int page_enabled() const {
		return (m_flags & SK3DROM_PAGE_ENABLED_BIT);}
	void set_page_enabled( const bool b) {
		if( b) m_flags |= SK3DROM_PAGE_ENABLED_BIT;
		else m_flags &= ~SK3DROM_PAGE_ENABLED_BIT;}

	void add_face( const sk3DRomTextureFace &f) {
		add_face( f.fid(), f.v( 0), f.v( 1), f.v( 2));
	}
	void add_face( const int fid, const rgVector a, const rgVector b,
		const rgVector c) {
		const int id = face_in_texture_page( fid);
		if( id >= 0) {
			m_face[ id] = sk3DRomTextureFace( fid, a, b, c);
			return;
		}
		m_face.push_back( sk3DRomTextureFace( fid, a, b, c));
	}

	void delete_face( const int id) {
		if( (id < 0) || (id >= face_count())) return;
		m_face.erase( m_face.begin() + id);
		clear_selected_vertex();
		clear_selected_vertices();
	}

	int face_in_texture_page( const int fid) const {
		for( int i = 0; i < int( m_face.size()); i++)
			if( fid == m_face[i].fid()) return i;
		return -1;
	}

	const char *page_name() const { return m_name;}

	void set_page_name( const char *p_name) {
		if( !p_name) return;
		strncpy( m_name, p_name, SK3DROM_TEX_PAGE_NAME_LEN - 1);
	}

	void update_face( const int fid, const rgVector a, const rgVector b,
		const rgVector c) {
		if( (fid < 0) || (fid >= int( m_face.size()))) return;
		m_face[ fid].update_vertices( a, b, c);
	}

	void change_face_fid( const int id, const int fid) {
		if( (id < 0) || (id >= int( m_face.size()))) return;
		// FIXME: Check if new "fid" will cause two faces to have the
		//	same "fid". Every face MUST have a unique fid.
		m_face[ id].set_fid( fid);
	}

	int select_face( const rgVector p);

	const sk3DRomTextureFace *face( const int id) const {
		if( (id < 0) || (id >= int( m_face.size()))) return 0;
		return &(m_face[id]);
	}

	void render_faces() const {
		for( int i = 0; i < int( m_face.size()); i++) {
			glVertex( m_face[i].v(0));
			glVertex( m_face[i].v(1));
			glVertex( m_face[i].v(2));
		}
	}

	void render_face( const int fid) const {
		if( (fid < 0) || (fid >= int( m_face.size()))) return;
		glVertex( m_face[ fid].v( 0));
		glVertex( m_face[ fid].v( 1));
		glVertex( m_face[ fid].v( 2));
	}

	void render_selected_vertices() const {
		for( int i = 0; i < int( m_sel_verts.size()); i++) {
			glVertex( m_face[ m_sel_verts[i].fid].v(
				m_sel_verts[i].vid));
		}
	}

	void transform_selected_vertices( const rgMatrix &m) {
		for( int i = 0; i < int( m_sel_verts.size()); i++) {
			m_face[ m_sel_verts[i].fid].transform_vertex(
				m_sel_verts[i].vid, m);
		}
	}

	void fuse_selected_vertices() {
		for( int i = 1; i < int( m_sel_verts.size()); i++) {
			m_face[ m_sel_verts[i].fid].update_vertex(
				m_sel_verts[i].vid,
				m_face[ m_sel_verts[0].fid].v(
					m_sel_verts[0].vid));
		}
	}

	void clear_selected_vertex() {
		m_sel_vertex = sk3DRomTexturePageSelVertex();
	}

	void clear_selected_vertices() { m_sel_verts.clear();}
	int selected_vertex_count() const { return int( m_sel_verts.size());}
	void select_vertex( const rgVector p, const float radius);

	bool vertex_is_selected() const {
		if( m_sel_vertex.fid >= 0) return true;
		return false;
	}

	rgVector selected_vertex() const {
		if( m_sel_vertex.fid < 0) return rgVector();
		return m_face[ m_sel_vertex.fid].v( m_sel_vertex.vid);
	}

	int selected_vertex_fid() const { return m_sel_vertex.fid;}
	int selected_vertex_vid() const { return m_sel_vertex.vid;}

	void update_selected_vertex( const rgVector v) {
		if( m_sel_vertex.fid < 0) return;
		m_face[ m_sel_vertex.fid].update_vertex( m_sel_vertex.vid, v);
	}

	void select_vertices( const rgVector n, const rgVector *v,
		const int count);
	void select_all_vertices();

	void bind_texture( const int tid) { m_texid = tid;}
	int texid() const { return m_texid;}

	int face_count() const { return int( m_face.size());}

	int flags() const { return m_flags;}
	int tex_unit() const { return m_tex_unit;}
	float uscale() const { return m_uscale;}
	float vscale() const { return m_vscale;}

	float set_uscale( float k) { return m_uscale = k;}
	float set_vscale( float k) { return m_vscale = k;}

	int read_versionA_page( std::ifstream &inf);

	void scale_selected_to_unit_texture();
	void scale_selected_to_unit_texture( const float p);
	void scale_selected_to_unit_texture( const float p,
		const rgVector vmin, const rgVector vmax);
	void scale_selected( const float t);
	void scale_selected( const float t, const rgVector vmin);
	void scale_selected( const float u, const float v);
	void scale_selected( const float u, const float v,
		const rgVector vmin);

	rgVector get_center_of_selected_vertices() const;
	void get_sel_verts_bbox( rgVector &vmin, rgVector &vmax) const;
	void left_justify_selected( const rgVector p);
	void left_justify_selected( const rgVector p, const rgVector vmin);
	void right_justify_selected( const rgVector p);
	void top_justify_selected( const rgVector p);
	void bottom_justify_selected( const rgVector p);
	void bottom_justify_selected( const rgVector p, const rgVector vmin);

	void composite_page( const sk3DRomTexturePage &src);
	int generate_stencil( rgBMP &bmp, const int dw, const int dh) const;

private:
	char m_name[ SK3DROM_TEX_PAGE_NAME_LEN];
	int m_flags;
	int m_tex_unit;
	float m_uscale;
	float m_vscale;
	int m_texid;
	std::vector <sk3DRomTextureFace> m_face;

	SelectionCache tri_sel_cache;

	sk3DRomTexturePageSelVertex m_sel_vertex;
	std::vector <sk3DRomTexturePageSelVertex> m_sel_verts;

	int read_versionA_faces( std::ifstream &inf);
};

struct sk3DRomBoneLink
{
	short bid;	// Which bone
	char jid;	// Which joint (0 = A, 1 = B)
};

#define SK3DROM_BONE_VERSION_A	1
#define SK3DROM_BONE_NAME_LEN	64
class sk3DRomBone
{
public:
	char bname[ SK3DROM_BONE_NAME_LEN];
	char pj;				// Pivot joint (0 = a, 1 = b)

	// Defines bones dimensions and default orientation
	rgVector a;				//
	rgVector b;				//
	rgVector ja;
	rgVector jb;

	short mid;				// Master bone id
	std::vector <sk3DRomBoneLink> link;	// Slave bones
	std::vector <short> vid;		// Index of controled vertices
	std::vector <short> mlist;		// Master transformation list

	/////////////////////////////////////////////////////////////////////
	// Bone coordinate systems.
	/////////////////////////////////////////////////////////////////////
	rgCoordSystem jcs;	// Joint coordinate system. The joint coord
				// system is used for limiting bone rotations.
				// This coordinate system does not affect the
				// positions of vertices or slave bones.
	rgCoordSystem fcs;	// Full (cached) coordinate system. This stores
				// the bones full coordinate system for a given
				// skeleton frame. If this bone is a root bone,
				// then "fcs" is equal to "jcs". "fcs" is set
				// by skeletal transformation functions.

	sk3DRomBone();

	const char *name() const { return bname;}
	const char *set_name( const char *p_name) {
		strncpy( bname, p_name, SK3DROM_BONE_NAME_LEN - 1);
		bname[ SK3DROM_BONE_NAME_LEN - 1] = '\0';
		return name();
	}

	rgVector get_pivot_joint() const;
	rgVector get_rotate_joint() const;

	rgVector get_jcs_pivot() const {
		if( pj <= 0) return ja;
		return jb;
	}

	rgVector get_jcs_rotate() const {
		if( pj <= 0) return jb;
		return ja;
	}

};

#define SK3DROM_BONE_FRAME_VERSION_A	1
class sk3DRomBoneFrame
{
public:
	rgCoordSystem rcs;	// Rotating coordinate system. This coord
				// system is used by all controlled vertices
				// and slave bones. This coordinate system
				// acts as the master coordinate system for
				// slave bones.

	int msb;	// Mounted Slave Bone
	int mmb;	// Master Mount Bone
	int cms;	// Chain Mounted Slave

	sk3DRomBoneFrame();
};

class sk3DRomSkelFrame
{
public:
	// The name for the skeleton frame
	char fname[SK3DROM_SKEL_FRAME_NAME_LEN];

	std::vector <sk3DRomBoneFrame> b;
	void *edata;	// "Hook" for extra data (can be used for anything).

	sk3DRomSkelFrame();

	const char *get_name() const { return fname;}
	void set_name( const char *p_name) {
		if( !p_name) {
			memset( fname, 0, SK3DROM_SKEL_FRAME_NAME_LEN);
			return;
		}
		strncpy( fname, p_name, SK3DROM_SKEL_FRAME_NAME_LEN - 1);
		fname[ SK3DROM_SKEL_FRAME_NAME_LEN - 1] = '\0';
	}

	//void update_skel_fcs( sk3DRomDrawContext &dc) const;
};


#define SK3DROM_ANI_CODE_TRANSITION	0
#define SK3DROM_ANI_CODE_STOP		1
#define SK3DROM_ANI_CODE_JUMP		2

#define SK3DROM_ANI_FRAME_FDATA_LEN	24
#define SK3DROM_ANI_FRAME_VERSION_A	1
class sk3DRomAniFrame
{
public:
	// The following are for extra data storage. The actual format of the
	// data contained within is dependant on the frame code. The extra
	// data blocks have been grouped into 6 DWORD's to provide convinience
	// in byte-ordering. However, a string of length 24 bytes can even
	// be stored in the extra data if desired.
	int data[6];

	sk3DRomAniFrame();
	int fid() const { return m_fid;}
	unsigned short time() const { return m_time;}
	unsigned short code() const { return m_code;}

	int set_fid( const int p_fid) { return (m_fid = p_fid);}
	unsigned short set_time( const unsigned short p_time) {
		return (m_time = p_time);}
	unsigned short set_code( const unsigned short p_code) {
		return (m_code = p_code);}

private:
	int m_fid;		// Frame used by this animation.
	unsigned short m_time;	// Display length (in milliseconds) of frame
	unsigned short m_code;	// The frames animation code.
};

// XXX: Helper class for setting/getting the extra data area of
//	sk3DRomAniFrame's when they contain the "jump" frame code.
#define SK3DROM_ANI_LOCK_TYPE_NONE	0
#define SK3DROM_ANI_LOCK_TYPE_JOINT	1
#define SK3DROM_ANI_LOCK_TYPE_BONE	2
#define SK3DROM_ANI_LOCK_TYPE_YAXIS	3
class sk3DRomAniFrame_jump
{
public:
	static void init_data( sk3DRomAniFrame &f) {
		set_times_to_skip( f, 0);
		set_probability( f, 1.0);
		set_delay( f, 0);
		set_des_sequence( f, 0);
		set_des_frame( f, 0);
		set_lock_type( f, SK3DROM_ANI_LOCK_TYPE_JOINT);
		set_target_bone( f, 0);
	}

	static void reset_skip_count( sk3DRomAniFrame &f) {
		f.data[0] &= 0x7FFF;}
	static void increment_skip_count( sk3DRomAniFrame &f) {
		f.data[0] = ((get_times_already_skipped( f) + 1) << 16) |
			get_times_to_skip( f);}
	static int get_times_already_skipped( const sk3DRomAniFrame &f) {
		return ((f.data[0] >> 16) & 0x7FFF);}

	static int set_times_to_skip( sk3DRomAniFrame &f, const int n) {
		return (f.data[0] = n & 0x7FFF);}
	static float set_probability( sk3DRomAniFrame &f, const float p) {
		return (*((float *)&(f.data[1])) = p);}
	static int set_delay( sk3DRomAniFrame &f, const int n) {
		return (f.data[2] = n);}
	static int set_des_sequence( sk3DRomAniFrame &f, const int sid) {
		return ((f.data[3] = (sid << 16) | (f.data[3] & 0xFFFF))
			>> 16) & 0xFFFF;
	}
	static int set_des_frame( sk3DRomAniFrame &f, const int fid) {
		return (f.data[3] = (f.data[3] & 0xFFFF0000) | (fid & 0xFFFF))
			& 0xFFFF;
	}
	static int set_lock_type( sk3DRomAniFrame &f, const int n) {
		return ((f.data[4] = (n << 16) | (f.data[4] & 0xFFFF))
			>> 16) & 0xFFFF;
	}
	static int set_target_bone( sk3DRomAniFrame &f, const int bid) {
		return (f.data[4] = (f.data[4] & 0xFFFF0000) | (bid & 0xFFFF))
			& 0xFFFF;
	}

	static int get_times_to_skip( const sk3DRomAniFrame &f) {
		return (f.data[0] & 0xFFFF);}
	static float get_probability( const sk3DRomAniFrame &f) {
		return *((float *)&(f.data[1]));}
	static int get_delay( const sk3DRomAniFrame &f) {
		return f.data[2];}
	static int get_des_sequence( const sk3DRomAniFrame &f) {
		return ((f.data[3] >> 16) & 0xFFFF);}
	static int get_des_frame( const sk3DRomAniFrame &f) {
		return (f.data[3] & 0xFFFF);}
	static int get_lock_type( const sk3DRomAniFrame &f) {
		return ((f.data[4] >> 16) & 0xFFFF);}
	static int get_target_bone( const sk3DRomAniFrame &f) {
		return (f.data[4] & 0xFFFF);}
};

class sk3DRomSequence
{
public:
	// Animations are an array of frames.
	std::vector <sk3DRomAniFrame> frame;

	void *edata;	// "Hook" for extra data (can be used for anything).

	sk3DRomSequence();

	int get_frame_with_fid( const int fid) const {
		for( int i = 0; i < int( frame.size()); i++)
			if( frame[i].fid() == fid) return i;
		return -1;
	}

	void delete_all_frames_with_fid( const int fid) {
		int i = 0;
		while( i < frame_count()) {
			if( (frame[i]).fid() == fid) delete_frame( i);
			else i++;
		}
	}

	void delete_frame( const int id) {
		if( (id < 0) || (id >= int( frame.size()))) return;
		frame.erase( frame.begin() + id);
	}

	const char *get_name() const { return m_name;}
	const char *set_name( const char *p_name) {
		strncpy( m_name, p_name, SK3DROM_SEQUENCE_NAME_LEN - 1);
		m_name[ SK3DROM_SEQUENCE_NAME_LEN - 1] = '\0';
		return m_name;
	}

	int frame_count() const { return int( frame.size());}

private:
	char m_name[ SK3DROM_SEQUENCE_NAME_LEN];
};

#define SK3DROM_NO_LOCK_SET		0
#define SK3DROM_GENERIC_LOCK_SET	1
class sk3DRomDrawContext
{
public:
	bool m_use_vertex_array;

	int m_cur_seq;			// The current sequence being animated
	int m_cur_frame;		// The current frame being animated
	int m_frame_start;
	int m_frame_stop;
	int m_frame_cur_time;
	int m_frame_total_time;

	sk3DRomSkelFrame m_skout;
	sk3DRomSkelFrame m_skold;	// Old skeleton frame (for animation)

	sk3DRomAniFrame m_old_frame;
	rgCoordSystem m_lock_cs;

	void pre_scale( const rgVector s) { m_scale.scale( s);}
	void pre_translate( const rgVector t) { m_translate.translate( t);}
	void pre_rotate( const float ang, const rgVector axis) {
		rgMatrix m;
		m.rotate( ang, axis);
		m_rotate = m * m_rotate;
	}

	void scale( const rgVector s) {
		rgMatrix m;
		m.scale( s);
		m_mat = m * m_mat;
	}
	void translate( const rgVector t) {
		rgMatrix m;
		m.translate( t);
		m_mat = m * m_mat;
	}
	void rotate( const float ang, const rgVector r) {
		// XXX: Always pivot on the root master pivot point.
		if( m_skout.b.empty()) return;
		rotate( ang, r,
			(mat()).pointMul( m_skout.b[ m_root_master].rcs.p));
	}

	void rotate( const float ang, const rgVector r, const rgVector p_pivot)
	{
		rgMatrix m;
		m.translate( p_pivot);
		m.rotate( ang, r);
		m.translate( -p_pivot);
		m_mat = m * m_mat;
	}

//	void update_matrix() { m_mat = (m_translate * m_rotate) * m_scale;}


	sk3DRomDrawContext();
	void gen_animation_frame( sk3DRomSkel &skel,
		sk3DRomSkelFrame &dc_skout);
	void gen_animation_frame( const float w, sk3DRomSkel &skel,
		sk3DRomSkelFrame &dc_skout);
	void init_current_frame( sk3DRomSkel &skel);
	void choose_next_frame( sk3DRomSkel &skel, sk3DRomSkelFrame &dc_skout);
	void choose_next_jump_frame( sk3DRomAniFrame *f, sk3DRomSkel &skel);
	void set_skeletal_lock( const sk3DRomAniFrame *f, sk3DRomSkel &skel,
		sk3DRomSkelFrame &dc_skout);
	void apply_skeletal_lock( const sk3DRomAniFrame *f,
		sk3DRomSkelFrame &s, sk3DRomSkel &skel);

	void init_fcs( const sk3DRomSkel &skel);

	int gen_frame_fcs( const sk3DRomSkel &s, const sk3DRomSkelFrame &f);

	void set_fcs( const int i, const rgCoordSystem &cs) {
		if( (i < 0) || (i >= m_fcs.size())) return;
		m_fcs.data[i] = cs;
	}

	rgCoordSystem get_fcs( const int i) const {
		if( (i < 0) || (i >= m_fcs.size())) return rgCoordSystem();
		return m_fcs.data[i];
	}

	void backup_matrices() {
		m_mat_backup = m_mat;
		m_scale_backup = m_scale;
		m_translate_backup = m_translate;
		m_rotate_backup = m_rotate;
		m_frame_transform_backup = m_frame_transform;
	}

	void restore_matrices() {
		m_mat = m_mat_backup;
		m_scale = m_scale_backup;
		m_translate = m_translate_backup;
		m_rotate = m_rotate_backup;
		m_frame_transform = m_frame_transform_backup;
	}

	void backup_main_skel() {
		backup_matrices();
		m_skel_backup = m_skout;
	}
	void restore_main_skel() {
		restore_matrices();
		m_skout = m_skel_backup;
	}

	rgCoordSystem get_full_bone_rcs(
		const sk3DRomSkel &s,
		const int bid,
		const sk3DRomSkelFrame &sf,
		rgCoordSystem &mcs) const;

	void reset_orientation() { m_orientation = rgCoordSystem();}

	void set_root_master( int rmid) {
		if( rmid < 0) rmid = 0;
		m_root_master = rmid;
	}

	rgMatrix m_scale;
	rgMatrix m_translate;
	rgMatrix m_rotate;
	rgMatrix mat() const;
	void reset_all_matrices() {
		m_frame_transform.loadIdentity();
		m_mat.loadIdentity();
		m_translate.loadIdentity();
		m_rotate.loadIdentity();
		m_scale.loadIdentity();
	}

protected:

	int m_root_master;
	rgMatrix m_mat;
	rgMatrix m_frame_transform;

	rgMatrix m_mat_backup;
	rgMatrix m_scale_backup;
	rgMatrix m_translate_backup;
	rgMatrix m_rotate_backup;
	rgMatrix m_frame_transform_backup;
	sk3DRomSkelFrame m_skel_backup;

//	rgMatrix m_frame_rotate;

//	rgVector m_pivot;
//	rgVector m_frame_pivot;

	int m_lock_type_set;

	rgCoordSystem m_orientation;
	rgCoordSystem m_prev_orientation;

	rgTempArray <rgCoordSystem> m_fcs;

	sk3DRomSkel *m_skel;

	void apply_frame_transformation() {
		m_mat = m_frame_transform * m_mat;
		m_frame_transform.loadIdentity();
	}

private:
	sk3DRomDrawContext( const sk3DRomDrawContext &);
	sk3DRomDrawContext &operator =( const sk3DRomDrawContext &);
};


class sk3DRomSkel
{
public:
	friend class sk3DRomModel;
	friend class sk3DRomDrawContext;

	std::vector <sk3DRomBone> b;
	std::vector <int> rb;	// Array of root bones

	// List of skeleton key frames
	std::vector <sk3DRomSkelFrame> sf;

	// Output skeleton for transformation functions.
	sk3DRomSkelFrame skout;

	sk3DRomSkel();

	int bone_count() const { return int( b.size());}
	int frame_count() const { return int( sf.size());}
	int sequence_count() const { return int( seq.size());}

	void init_skel_frame( sk3DRomSkelFrame &p_sf);

	void update_fcs( sk3DRomDrawContext &dc, const sk3DRomSkelFrame &sf)
		const;

	int get_total_seq_time( const int sid);
	void get_seq_frame_and_scalar( const int sid,
		const int ttime, const float p, int &fid, float &s);

	void inbetween_skeleton( const float w,
		const sk3DRomSkelFrame &ska, const sk3DRomSkelFrame &skb);
	void inbetween_skeleton( const float w, const sk3DRomSkelFrame &ska,
		const sk3DRomSkelFrame &skb, sk3DRomSkelFrame &skc);

	void copy_mount_data( sk3DRomSkelFrame &des,
		const sk3DRomSkelFrame &src) const;

	void get_bone_coords( const short bid, const short sid, rgVector &a,
		rgVector &b);
	void get_bone_coords( const short bid, const sk3DRomSkelFrame &p_sf,
		rgVector &a, rgVector &b);

	int get_root_master( const int bid, const sk3DRomSkelFrame &s) const;

	rgCoordSystem get_full_bone_jcs( const short bid) const;
	// FIXME: Remove obsolete "get_full_bone_rcs" functions.
	rgCoordSystem get_full_bone_rcs( const short bid, const short sfid)
		const {
		return get_full_bone_rcs( bid, sf[sfid]);
	}
	rgCoordSystem get_full_bone_rcs( const int bid,
		const sk3DRomSkelFrame &s) const {
		rgCoordSystem mcs;
		return get_full_bone_rcs( bid, s, mcs);};
	rgCoordSystem get_full_bone_rcs( const int bid,
		const sk3DRomSkelFrame &s, rgCoordSystem &mcs) const;

	void get_full_bone_rcs( const short bid, const short sfid,
		rgCoordSystem &rcs) const {
		rcs = get_full_bone_rcs( bid, sfid);
	}

	void rotate_rcs( const short bid, const short sid, const float ang,
		const rgVector axis);
	void rotate_rcs( const short bid, sk3DRomSkelFrame &p_sf,
		const float ang, const rgVector axis);
	void translate_rcs( const int bid, sk3DRomSkelFrame &p_sf,
		const rgVector v) const;

	void reset_bone( const short bid, const short sid);
	void reset_bone( const short bid,
		sk3DRomSkelFrame &p_sf) {
		if( (bid < 0) || (bid >= int( b.size()))) return;
		p_sf.b[bid].rcs = b[bid].jcs;};

	void reset_bone_and_slaves( const short bid, const short sid);
	void reset_bone_and_slaves( const short bid,
		sk3DRomSkelFrame &p_sf);

	void build_root_bone_list();

	void build_master_list();
	void build_bone_master_list( const int bid);

	// This only needs to be once, and then, only for skeletons loaded
	// from a sk3DROM_MODEC (or lower) sk3DRom file.
	void bind_skeleton();

	void bind_bone_to_bone( const int mid, const int bid);

	// Description: Unbinds a given bone (making it a root bone). All
	//	skeleton frames and "skout" are also updated to reflect the
	//	unbinding.
	void unbind_bone( const int bid);

	// Description: Mounts bone "bid" onto bone "tid" for skeleton frame
	//	"p_sf".
	void mount_bone( const int tid, const int bid,
		sk3DRomSkelFrame &p_sf);

	// Description: Unmounts a mounted slave bone (causing bone "bid" to
	//	become "free"). The unmounted slave will still be a master of
	//	its own slaves (if it had any) however.
	void unmount_bone( const int bid, sk3DRomSkelFrame &p_sf);

	// Description: Unmounts all slaves of master bone "bid". The master
	//	bone will no longer have any more slaves, but it will still
	//	remain a slave itself (if it was a slave before).
	void unmount_bone_slaves( const int bid, sk3DRomSkelFrame &p_sf);

	void reset_skel_frame( const int sid);
	void reset_all_skel_frames();
	
	void gen_absolute_joint_verts();

	void delete_skeleton_frame( const int sid);

	sk3DRomSequence &add_sequence() {
		seq.push_back( sk3DRomSequence());
		return seq[ seq.size() - 1];
	}

	void delete_sequence( const int sid);
	void delete_all_sequences() {
		// XXX: It is most efficient to delete from the back.
		while( !(seq.empty())) delete_sequence( sequence_count() - 1);
	}

	sk3DRomSkelFrame &animate( const int p_elapsed);
	sk3DRomSkelFrame &animate( const int sid, const int fid,
		const int p_time, const int p_elapsed);
	sk3DRomSkelFrame &animate( const int p_elapsed,
		sk3DRomDrawContext *dc);
	sk3DRomSkelFrame &animate( const int sid, const int fid,
		const int p_time, const int p_elapsed,
		sk3DRomDrawContext *dc);

	// m_draw_context is provided for convinience if only one client will
	// be using the skeleton. If the skeleton will be used by multiple
	// clients, then each client should have their own draw context.
	sk3DRomDrawContext m_draw_context;


protected:
	// List of skeleton animation sequences.
	std::vector <sk3DRomSequence> seq;

	void notify_sequences_of_frame_delete( const int fid);
	void notify_sequences_of_frame_swap( const int old_fid,
		const int new_fid);

	void gen_mount_order_root_bone_list(
		std::vector <int> &mlist, const sk3DRomSkelFrame &p_sf);
	void get_bone_slave_list( const int bid, std::vector <int> &blist);
	void get_bone_mount_slave_list( const int bid,
		const sk3DRomSkelFrame &p_sf, std::vector <int> &blist);
	void set_pivot_joint( const int bid, const short jid);
	void bind_rcs_to_cs( rgCoordSystem &rcs, rgCoordSystem &cs) const {
		rcs = cs.change_csys_csys( rcs);};
	void create_bone_jcs( sk3DRomBone &b);
	void bind_bone_to_jcs( rgCoordSystem *mcs, sk3DRomBone &bn);
	void unbind_bone_jcs( const int bid);
	void add_root_bone( const int bid);
	void remove_root_bone( const int bid);
	void add_bone_link( sk3DRomBone &p_b, const int link);
	void remove_bone_link( sk3DRomBone &p_b, const int link);
	void bind_slaves( const short mid);
	bool bone_id_is_invalid( const int bid) const {
		if( (bid < 0) || (bid >= int( b.size()))) return true;
		return false;};

	bool bone_is_in_mount_chain( const int bid, const int cid,
		sk3DRomSkelFrame &p_sf);

	void position_bone( sk3DRomBone &b, const rgVector ja,
		const rgVector jb);

	void change_skel_mount_bone_id( const int old_bid,
		const int new_bid, sk3DRomSkelFrame &p_sf);
	void notify_mounts_of_bone_id_change( const int old_bid,
		const int new_bid);

	void unmount_from_entire_skeleton( const int bid);
	void unmount_everything( const int bid, sk3DRomSkelFrame &p_sf);
	void unmount_slaves( const int bid, sk3DRomSkelFrame &p_sf);
	void unmount_from_master( const int bid, sk3DRomSkelFrame &p_sf);

	// Description: "Unbinds" slave bones, causeing them to beomce root
	//	bones. Thus all  unbound slave bones must have their Joint
	// Coordinate System (jcs) be recreated.
	void unbind_bone_slaves( sk3DRomBone &p_bn);// {
	//	while( !(p_bn.link.empty())) unbind_bone( p_bn.link[0].bid);
	//}
};

class sk3DRomVertexArray_RenderGroup
{
public:
	int m_start;
	int m_count;
	GLuint m_texid;

	sk3DRomVertexArray_RenderGroup() {
		m_start = 0;
		m_count = 0;
		m_texid = 0;
	}

	sk3DRomVertexArray_RenderGroup( int p_start) {
		m_start = p_start;
		m_count = 0;
		m_texid = 0;
	}
};

// Vertex array rendering facility.
// This is provided for convinience.
class sk3DRomVertexArray
{
public:
	sk3DRomVertexArray() {
		m_index = 0;
	}
	virtual ~sk3DRomVertexArray() {
		if( m_index) delete [] m_index;
	}

	virtual int create_arrays( const sk3DRomModel &mdl) = 0;
	virtual void update_arrays( const sk3DRomModel &mdl) = 0;
	virtual void render_arrays() = 0;
	virtual void delete_arrays() = 0;

	int non_alpha_face_count( const sk3DRomModel &mdl);
	void non_alpha_faces( const sk3DRomModel &mdl, std::vector <int> &naf);

	int create_face_index_array( const sk3DRomModel &mdl) {
		if( m_index) { delete [] m_index; m_index = 0;}
		m_index = create_face_index_array( mdl, m_face_count,
			m_render_group);
		if( !m_index) return -1;
		return 0;
	}

	const std::vector <sk3DRomVertexArray_RenderGroup> &
		render_group() const { return m_render_group;}
	int face_count() const { return m_face_count;}
	const unsigned short *index_array() const { return m_index;}

private:
	unsigned short *m_index;
	int m_face_count;
	std::vector <sk3DRomVertexArray_RenderGroup> m_render_group;

	// Description: Creates an returns an array containing face-vertex
	//	index values. The faces are sorted by texture usage. Also,
	//	"rgrp" is populated with a list of "render-groups". These
	//	indiciate which faces should be rendered together for optimal
	//	speed.
	unsigned short *create_face_index_array( const sk3DRomModel &mdl,
		int &count,
		std::vector <sk3DRomVertexArray_RenderGroup> &rgrp);
};

#define SK3DROM_RENDER_FORMAT_V4N3E1T2E6	0
class sk3DRomModel
{
public:

	// List of vertices for all models
	std::vector <sk3DRomVertex> v;

	// List of vertices not controlled by any bones.
	// If there are no bones at all, then bone_less_verts will be NULL
	// and sk3DRomSkel functions still know that all vertices are
	// bone-less (this is because they first check if there any bones
	// at all). If there are bones, then bone_less_verts can be NULL
	// if all vertices have controlling bones, or it can be as long
	// as the total number of bone-less vertices.
	std::vector <int> bone_less_verts;


	// List of faces for all models
	std::vector <sk3DRomFace> f;

	// List of textures for the model
	std::vector <sk3DRomTexture> texture;

	// List of groups
	std::vector <sk3DRomGroup> group;

	// Main skeleton
	sk3DRomSkel skel;

	rgCSGBrush *m_brush_cursor;
	rgCSGWorld *m_csg;

	// "bsp_tree" points to an array of BSP trees. BSP tree's are used to
	// store face's that make use of the alpha component (transparent and
	// translucent faces). This is because face's which use an alpha
	// component must be rendered in back to front order.
	std::vector <rgBSPTree <rgBSP_NTTC_ExtData> *> bsp_tree;
	std::vector <int> bsp_bone_id;

	// Description: Before any 3DRom data is saved at all, "pre_save" is
	//	called to perform final processing before saving of 3DRom data
	//	begins.
	int (*pre_save)( sk3DRomModel *mdl);

	// Description: Call-back function for writing out custom data to
	//	a 3DRom file. When all other 3DRom data has been written to
	//	the file, this function will be called. If the function has no
	//	custom data to write out, then it should return without
	//	writing to the file. If it does have custom data to write, then
	//	it MUST write it within sk3DROM_CUSTOM tags. There is no limit
	//	to the number of sk3DROM_CUSTOM tags that can be written out.
	int (*write_custom_data)( std::ostream &des, sk3DRomModel *mdl);

	// Description: Call-back function for reading custom data from a
	//	3DRom file. When a top level custom tag (not custom tags
	//	contained within other custom tags) is encountered, this
	//	function will be called to process it. If the function does not
	//	understand the custom tag, then it should return reading
	//	nothing (sk3DRomModel will then automatically skip the custom
	//	tag).
	//
	//	NOTE: sk3DRomModel does not save any information about custom
	//	tags when it loads 3DRom files. If a file contains one or more
	//	custom tags, it is the responsibility of the "read_custom_data"
	//	function to save all custom data (some where) so that a
	//	subsequent save operation can write the custom data as well.
	//	If it is not possible to save the custom data information, then
	//	do not modify the file which contained the custom data, instead
	//	save to a different file.
	//
	//	sk3DRomModel provides the function "store_custom_data" which
	//	will write (or append) a custom tag to a file. When saving, use
	//	"include_custom_data" to add stored custom data to the file
	//	being saved.
	//
	//	Custom Tag Ordering:
	//
	//		- A 3DRom file can only contain at most ONE top level
	//		  custom tag. It is an error to have more than one.
	//
	//		  1st level custom tags are ONLY read and written by
	//		  sk3DRomModel. sk3DRomModel does not read or write
	//		  custom tags in any other level (that is, it does not
	//		  interpret the custom data within custom tags).
	//
	//
	//		- 2nd level custom tags are order independent.
	//		  This means that multiple 2nd level tags can be in any
	//		  order. A top level custom tag often contains 2nd
	//		  level custom tag from different applications which
	//		  can reorder 2nd level tags as they like. However
	//		  data within a 2nd level tag will never be modified
	//		  by applications which do not understand them. There-
	//		  fore order dependent data can be stored in a single
	//		  2nd level tag and it is gaurenteed that its data will
	//		  be unmodified.
	//
	//		  2nd level custom tags are read and written by
	//		  3rd party applications. A 3rd party application
	//		  must never modify the data within a 2nd level
	//		  custom tag unless it understands it.
	//
	//		  In other words, 2nd level custom tags are 3rd party
	//		  custom data blocks.
	//
	int (*read_custom_data)( std::ifstream &inf, sk3DRomModel *mdl,
		const int seek_end);

	// Description: Stores custom tags.
	//int store_custom_tag( std::fstream &des, std::ifstream &src,
	//	sk3DRomTag &t);
	int store_custom_tag( std::iostream &des, std::ifstream &src,
		sk3DRomTag &t);

	int include_custom_tag( std::ostream &des, std::ifstream &src);

	sk3DRomModel();
	~sk3DRomModel();

	// Description: Produces a BSP tree storing all faces which make use
	//	of an alpha component (either from an alpha enabled texture
	//	or from the alpha colour component).
	//
	//	NOTE: This is a very slow computational and memory intensive
	//		function. This function is best used in a seperate
	//		process to generate the BSP tree's so that another
	//		process can use. This helps to conserve memory and
	//		negate the effects of memory leaks (if any).
	void compile_alpha_faces();

	void build_bone_less_list();

	void clear();
	void gen_master_joint_links();
	void gen_flat_shade_normals();
	void gen_smooth_shade_normals();

	// Description: Takes the texture data described in the model's texture
	//	pages and produces "render ready" texture data.
	void compile_texture_data();
	void compile_page_texture_data( const int pid);

	int load( const char *fname);


	// Description: Causes the model to reload all of its textures. The
	//	model uses the texture's absolute file location (given by
	//	the textures file name) if "paths" is NULL. Otherwise,
	//	the model tries to load the texture file names from the given
	//	path(s). If the texture can not be found in "paths" then
	//	the model WILL NOT try to load from the absolute file name
	//	locaiton. Multiple paths can be given in "paths" and the
	//	model will try them all until one works or non work. The
	//	last element in "paths" MUST be a NULL string/path.
	//
	//	NOTE: The last element in "paths" MUST be a NULL sting/path.
	//
	// Returns: 0 is returned if all textures were loaded without errors.
	//	If the model had no textures at all, then 0 is returned also.
	//	A negative value is returned if at least one texture failed
	//	to load correctly.
	int load_textures( const char **paths, const int count);

	int save( const char *fname);

	// Produces transformed vertices for the selected skeleton frame.
	// The version of "transform_skel_frame" which takes no parameters
	// simply copies the vertices to the "tv" array as they are (no
	// transformation is done).
	//
	// Providing no skeleton frame or index number. If a draw context
	// (sk3DRomDrawContext) is provided, "dc->m_skout" is choosen as the
	// skeleton frame to use, otherwise "rg3DRomModel::skel.skout"
	// (this->skel.skout) is used instead.
	void transform_skel_frame() { transform_skel_frame( 0, -1, 0);}
	void transform_skel_frame( const rgMatrix *m) {
		transform_skel_frame( m, -1, 0);}
	void transform_skel_frame( const rgMatrix *m, sk3DRomDrawContext *dc) {
		transform_skel_frame( m, -1, dc);}
	void transform_skel_frame( sk3DRomDrawContext *dc) {
		transform_skel_frame( 0, -1, dc);}

	// Provinding a skeleton frame index number.
	void transform_skel_frame( const rgMatrix *m, const int sfid) {
		transform_skel_frame( m, sfid, 0);}
	void transform_skel_frame( const int sfid) {
		transform_skel_frame( 0, sfid, 0);}
	void transform_skel_frame( const int sfid, sk3DRomDrawContext *dc) {
		transform_skel_frame( 0, sfid, dc);}
	void transform_skel_frame( const rgMatrix *m, int sfid,
		sk3DRomDrawContext *dc);

	// Providing an actual skeleton frame.
	void transform_skel_frame( const rgMatrix *m,
		const sk3DRomSkelFrame &s) {
		transform_skel_frame( m, s, 0);}
	void transform_skel_frame( const sk3DRomSkelFrame &s) {
		transform_skel_frame( 0, s, 0);}
	void transform_skel_frame( const sk3DRomSkelFrame &s,
		sk3DRomDrawContext *dc) {
		transform_skel_frame( 0, s, dc);}
	void transform_skel_frame( const rgMatrix *m,
		const sk3DRomSkelFrame &s, sk3DRomDrawContext *dc);

	void prepare_texture_coords_for_rendering();
	void prepare_normals_for_rendering();

	void render_verts();
	void render_verts( const int *vlist, const int count);
	void render_group_verts();
	void render_group_verts( const int *vlist, const int count);

	void render_single_face( bool use_normals, bool use_texture,
		const int i);
	void render_single_face( bool use_normals, bool use_texture,
		const int i, float *vc1, float *vc2, float *vc3);
	void render_faces( bool use_normals, bool use_textures);
	void render_faces( bool use_normals, bool use_textures,
		const int *flist, const int count);
	void render_faces( bool use_normals, bool use_textures,
		const int *flist, const int count, bool force_no_bsp);
	void render_faces( bool use_normals, bool use_texture,
		const int *flist, const int count, bool force_no_bsp,
		bool use_group_color);
	void render_faces( bool use_normals, bool use_textures,
		const int *flist, const int count, bool force_no_bsp,
		bool use_group_color,
		bool (*)( const sk3DRomModel &, const int));

	void render_bsp_faces( rgVector p, bool use_normals, bool use_textures,
		const int sid);
	void render_bsp_faces( rgVector p, bool use_normals, bool use_textures,
		const int sid, int *flist, const int count);
	void render_bsp_faces( const sk3DRomDrawContext *dc, rgVector p,
		bool use_normals, bool use_textures, int *flist,
		const int count);

	void render_skel_frame();
	void render_skel_frame( const int sid);
	void render_bone( const sk3DRomSkelFrame &sf, const int bid);
	void render_page( const int pid) const;
	void render_page_face( const int pid, const int fid) const {
		if( (pid < 0) || (pid >= int( m_page.size()))) return;
		m_page[ pid].render_face( fid);
	}

	void get_tv_bounding_box( float &left, float &right, float &bottom,
		float &top, float &front, float &back);

	void get_full_bone_transformation( const short bid, const short sfid,
		rgMatrix &m);

	// Model query functions.
	int vertex_count() const { return int( v.size());}
	int face_count() const { return int( f.size());}
	int group_count() const { return int( group.size());}
	int page_count() const { return int( m_page.size());}
	int texture_count() const { return int( texture.size());}
	int skeleton_count() const { return int( skel.sf.size());}
	int sequence_count() const { return skel.sequence_count();}
	int get_group_id( const char *gname) const;
	void set_face_group_color3( const int fid) const {
		if( (fid < 0) || (fid >= face_count())) return;
		if( (f[fid].gid < 0) || (f[fid].gid >= group_count()))
			glColor3( m_default_groupless_color);
		else group[ f[fid].gid].set_color3();
	}
	//rgVector face_group_color( const int fid) const {
	//	if( (fid < 0) || (fid >= face_count())) return rgVector();
	//	if( (f[fid].gid < 0) || (f[fid].gid >= group_count()))
	//		return rgVector();
	//	return group[ f[fid].gid].color();
	//}
	const sk3DRomTexturePage *page( const int pid) const {
		if( (pid < 0) || (pid >= page_count())) return 0;
		return &(m_page[ pid]);
	}
	const sk3DRomSequence *sequence( const int sid) const {
		if( (sid < 0) || (sid >= sequence_count())) return 0;
		return &(skel.seq[ sid]);
	}
	sk3DRomAniFrame *get_editable_ani_frame( const int sid, const int fid){
		if( (sid < 0) || (sid >= sequence_count())) return 0;
		if( (fid < 0) || (fid >= skel.seq[sid].frame_count()))
			return 0;
		return &(skel.seq[sid].frame[fid]);
	}
	bool face_hidden( const int fid) const {
		if( (fid < 0) || (fid >= face_count())) return false;
		if( v[ f[fid].vid[0]].hidden()) return true;
		if( v[ f[fid].vid[1]].hidden()) return true;
		if( v[ f[fid].vid[2]].hidden()) return true;
		if( (f[fid].gid < 0) || (f[fid].gid >= group_count()))
			return false;
		if( group[ f[fid].gid].hidden()) return true;
		return false;
	}
	rgVector face_normal( const int fid) const {
		if( (fid < 0) || (fid >= face_count())) return rgVector();
		return (rgCalcTriangleNormal(
			get_tv_vertex( f[fid].vid[0]),
			get_tv_vertex( f[fid].vid[1]),
			get_tv_vertex( f[fid].vid[2]))).normalize();
	}
	

	// Member access functions
	sk3DRomTexturePage *edit_page( const int pid) {
		if( (pid < 0) || (pid >= page_count())) return 0;
		return &(m_page[ pid]);
	}


	// Model editing functions.
	sk3DRomVertex *add_vertex( const rgVector p_v);
	sk3DRomFace *add_triangle( const int v1, const int v2, const int v3);
	sk3DRomGroup *add_group( const sk3DRomGroup &g);
	sk3DRomBone *add_bone( rgVector ja, rgVector jb);
	void add_skeleton_frame( const sk3DRomSkelFrame &s);
	sk3DRomSequence *add_sequence( const sk3DRomSequence &s) {
		return &((skel.add_sequence()) = s);
	}
	sk3DRomAniFrame *add_sequence_frame( const int sid) {
		if( (sid < 0) || (sid >= sequence_count())) return 0;
		skel.seq[sid].frame.push_back( sk3DRomAniFrame());
		return &(skel.seq[sid].frame[ skel.seq[sid].frame_count()
			- 1]);
	}
	int add_texture();
	int add_page();
	void add_tv_face_to_page( const int pid, const int fid,
		const rgVector a, const rgVector b, const rgVector c);
	void delete_vertex( const int vid);
	void delete_face( const int fid);
	void delete_group( const int gid);
	void delete_bone( const int bid);
	void delete_bone( const int bid, sk3DRomSkelFrame *sf);
	void delete_texture( const int tid);
	void delete_page( const int pid);
	void delete_all_skeleton_frames();
	void delete_sequence( const int sid) { skel.delete_sequence( sid);}
	void delete_all_sequences() { skel.delete_all_sequences();}
	void delete_sequence_frame( const int sid, const int fid) {
		if( (sid < 0) || (sid >= sequence_count())) return;
		(skel.seq[sid]).delete_frame( fid);
	}
	// FIXME: Spelling mistake for below function.
	void offset_tv_verticies( const int *vlist, int count,
		const rgVector p_v);
	void transform_tv_vertices( const rgMatrix &m) {
	        if( !tv) return;
	        for( int i = 0; i < vertex_count(); i++)
	                set_tv_vertex( i, m.pointMul( get_tv_vertex( i)));
	}
	void remove_vertex_from_group( const int vid, const int gid);
	void add_vertex_to_group( const int vid, const int gid);
	void hide_group( const int gid);
	void reveal_group( const int gid);
	int ortho_select_tv_triangle( rgVector p, const rgVector n);
	void ortho_select_bone( rgVector p, const rgVector n,
        	const sk3DRomSkelFrame &sf, int &bid, int &jid);
	void ortho_select_bone( rgVector p, const rgVector n,
		const sk3DRomSkelFrame &sf, int &bid, int &jid,
		const float min); 
	void bind_vertex_to_bone( const int vid, const int bid);
	void bind_vertices_to_bone( const int *vlist, int count,
		const int bid);
	void unbind_vertex( const int vid);
	void unbind_vertices( const int *vlist, int count);
	void bind_brush_to_bone( rgCSGBrush *b, const int bid);
	void unbind_brush( rgCSGBrush *b) const;
	void duplicate_vertices( const int *vlist, int count);
	void duplicate_mesh( const int *vlist, int count);
	void invert_mesh( const int *vlist, const int count);
	void scale_model( const float x, const float y, const float z);
	void mirror_tv_vertices( const int *vlist, int count,
		const rgVector origin, const rgVector normal,
		const rgVector mirror_vec);
	rgVector get_center_of_tv_vertex_cluster( const int *vlist, int count);
	void set_sequence_name( const int sid, const char *p_name) {
		if( (sid < 0) || (sid >= sequence_count())) return;
		(skel.seq[ sid]).set_name( p_name);
	}
	/*
	void set_sequence_frame_code( const int sid, const int fid,
		const int p_code) {
		if( (sid < 0) || (sid >= sequence_count())) return;
		if( (fid < 0) || (fid >= skel.seq[sid].frame_count())) return;
		(skel.seq[sid].frame[ fid]).set_code( p_code);
	}
	*/
	void set_page_name( const int pid, const char *p_name) {
		if( (pid < 0) || (pid >= page_count())) return;
		m_page[ pid].set_page_name( p_name);
	}
	void bind_texture_to_page( const int pid, const int tid) {
		if( (pid < 0) || (pid >= page_count())) return;
		m_page[ pid].bind_texture( tid);
	}
	int select_page_face( const int pid, const rgVector mp) {
		if( (pid < 0) || (pid >= page_count())) return -1;
		return m_page[ pid].select_face( mp);
	}

	void select_page_vertices( const int pid, const rgVector n,
		const rgVector *p_v, const int count) {
		if( (pid < 0) || (pid >= page_count())) return;
		m_page[ pid].select_vertices( n, p_v, count);
	}
	void transform_selected_page_vertices( const int pid,
		const rgMatrix &m) {
		if( (pid < 0) || (pid >= page_count())) return;
		m_page[ pid].transform_selected_vertices( m);
	}
	void clear_selected_page_vertex( const int pid) {
		if( (pid < 0) || (pid >= page_count())) return;
		m_page[ pid].clear_selected_vertex();
	}
	void clear_selected_page_vertices( const int pid) {
		if( (pid < 0) || (pid >= page_count())) return;
		m_page[ pid].clear_selected_vertices();
	}
	void update_page_face( const int pid, const int fid,
		const rgVector a, const rgVector b, const rgVector c) {
		if( (pid < 0) || (pid >= page_count())) return;
		m_page[ pid].update_face( fid, a, b, c);
	}
	void update_selected_page_vertex( const int pid, const rgVector p) {
		if( (pid < 0) || (pid >= page_count())) return;
		m_page[ pid].update_selected_vertex( p);
	}
	void select_page_vertex( const int pid, const rgVector p,
		const float radius) {
		if( (pid < 0) || (pid >= page_count())) return;
		m_page[ pid].select_vertex( p, radius);
	}
	void fuse_selected_page_vertices( const int pid) {
		if( (pid < 0) || (pid >= page_count())) return;
		m_page[ pid].fuse_selected_vertices();
	}
	void composite_page( const int des_pid, const int src_pid) {
		if( (des_pid < 0) || (des_pid >= page_count())) return;
		if( (src_pid < 0) || (src_pid >= page_count())) return;
		m_page[ des_pid].composite_page( m_page[ src_pid]);
	}
	void scale_selected_page_faces_to_unit_texture( const int pid,
		const float p_border);

	// Description: Repositions a bone, causing "ja" and "jb" to the
	//	new default joint locations of the bone. The bone will have its
	//	Joint Coordinate System (jcs) recomputed. Before the new jcs
	//	is created, all vertices and slave bones are unbound, then
	//	the jcs is created, then all vertices and slaves are rebound
	//	to the bone.
	void reposition_bone( const int bid, const rgVector ja,
		const rgVector jb);

	// Description: Returns absolute vertex locations of joints "a", "b"
	//	and the pivot joint identifier.
	int get_absolute_bone_vertices( const sk3DRomSkelFrame &sf,
		const int bid, rgVector &a, rgVector &b);

	// Description: Reveal's all vertices by setting their hidden bit flag
	//	to zero.
	void reveal_all_vertices();
	void set_vertex_gid( int *vlist, int count, const int gid);

	/* FIXME: "set_vertex" has a bad running time due to rebinding.
	// Description: Set's vertex "i" to be at position "p_v". If vertex
	//	"i" is bound to a bone, then it is "rebound" so that "p_v"
	//	is relative to owning bone's coordinate system. Use this
	//	function rather than setting the vertex array "v" manually
	//	(unless the owning bone is taken into consideration).
	void set_vertex( const int i, const rgVector p_v);
	*/

	void set_vertex_from_tv( const int *vlist, int count);
	void set_vertex_from_tv( const int *vlist, int count,
		const int sid);
	void set_vertex_from_tv( const int *vlist, int count,
		const sk3DRomSkelFrame &sf);
	void set_tv_vertex( const int i, const rgVector p_v);


	void auto_gen_joints();

	/*
	 * Description: Takes a bone orientation "a" and "b", for the bone
	 *	"bone", and determines the forward angle "fang" and the side
	 *	angle "sang".
	 */
	void get_bone_angle( rgVector a, rgVector b, sk3DRomBone &bone,
        	float &fang, float &sang);

	void map_verticies_to_skeleton();

	void set_pivot_joint( const int bid, const short jid);
	void swap_pivot_joint( const int bid);

	bool create_transformation_array();

	int get_render_format();

	rgVector get_tv_vertex( const int i) const; 
	void get_tv_face_verticies( const int i, rgVector *p_v);

	void clear_custom_tag_memory();
	int get_custom_tag_memory_size();

	void gen_planar_map( bool p_visible, const rgMatrix &m, const int pid);
	void gen_planar_map( bool p_visible, const rgCoordSystem &cs,
		const int pid);

	const int *get_faces_in_group( const int gid, int &count);
	void set_group_of_visible_faces( int gid);
	void set_color_of_face_group( const int gid, float r, float g,
		float b);
	void set_alpha_of_face_group( const int gid, float a);
	void force_flat_face_group( const int gid);
	void unforce_flat_face_group( const int gid);
	void set_bsp_state_for_face_group( const int gid, bool p_state);
	void set_use_color_of_face_group( const int gid);
	void unset_use_color_of_face_group( const int gid);

	const sk3DRomVertexArray *get_vertex_array() const {
		return m_gl_vertex_array;}
	int create_vertex_array();
	void update_vertex_array();
	void render_vertex_array( sk3DRomDrawContext *dc);
	void delete_vertex_array();

protected:
	sk3DRomVertexArray *m_gl_vertex_array;

	int *m_face_list;
	rgVector m_default_groupless_color;

	// FIXME: "tv" should only have ONE render format (since it is not
	//	designed for optimal rendering anyway).
	// Transformed vertices and other render data.
	// Default format is 64 bytes per vertex, consisting of vertex,
	// normal, texture coordinate, and extra data.
	rgFloat *tv;

	// FIXME: "tv" should only have one format, and fast_get_tv_vertex
	//	should use that format.
	friend class sk3DRomVNTArray;	// Circumvent protection for speed
	void fast_get_tv_vertex( const int i, float &x, float &y, float &z)
		const
	{
		const rgFloat *tv_ptr = tv + (i << 4);
		x = tv_ptr[0];
		y = tv_ptr[1];
		z = tv_ptr[2];
	}

	std::vector <sk3DRomTexturePage> m_page;

	int custom_tag_memory_length;
	char *custom_tag_memory_storage;
	int store_custom_tag_in_memory( std::ifstream &inf, const int len);

	int tv_array_element_count;

	// Default format is V4N3E1T2E6
	int render_format;

	SelectionCache tri_sel_cache;
	bool triangle_is_hidden( const int i) const {
		if( (v[ f[i].vid[0]].hidden()) || (v[ f[i].vid[1]].hidden()) ||
			(v[ f[i].vid[2]].hidden())) return true;
		return false;
	}

	//void scale_model_skeleton( const float x, const float y,
	//	const float z);

	bool load_texture_from_paths( sk3DRomTexture &t,
		const char **paths, int count) const;

	void add_vertex_to_bone_less_list( const int vid);
	void remove_vert_from_bone_less_list( const int vid);

	bool triangle_in_mesh( const int fid, const int *vlist,
		const int count, int *vid) const;

	// Description: Runs through every group and builds lists of vertices
	//	owned by each group.
	void build_group_vert_lists();
	void clear_bid_for_all_verts() {
        	for( int i = 0; i < int( v.size()); i++) v[i].m_bid = -1;}
	void set_bid_for_all_verts();

	void notify_verts_of_bone_swap( const int old_bid,
	        const int new_bid);
	void notify_skel_of_bone_id_change( const int old_bid,
		const int new_bid);

	void notify_pages_of_face_delete( const int fid);
	void notify_pages_of_face_swap( const int old_fid, const int new_fid);
	void notify_pages_of_texture_delete( const int tid);
	void notify_pages_of_texture_swap( const int old_tid,
		const int new_tid);

	void clear_bsp_data();

	void bind_sub_skel_verticies( const short bid);
	void unbind_sub_skel_verticies( const short bid);

	void bind_bone_verticies( const short bid);
	void unmap_bone_vertices( const int bid, sk3DRomSkelFrame *sf);
	void unbind_bone_vertices( const int bid, sk3DRomSkelFrame *sf);
	void unbind_vertex( const int vid, rgCoordSystem *rcs);


	void transform_bone_verticies_and_slaves( const short bid,
	        const short sfid, rgMatrix *p_m);

	void map_bone_verticies_and_slaves( const short bid,
	        const short sfid, rgCoordSystem *mcs);

	void transform_bone_verticies( const short bid, const short sfid,
		rgMatrix &p_m);

	void transform_bone_verticies( const short bid, const short sfid,
		rgCoordSystem &cs);

	friend int load_3DRom( const char *fname, sk3DRomModel &mdl);

	//************ Functions for checking corruption ************//
	void check_faces();


	//************ Reading sk3DRom Files ************//
	int read_face_data( std::ifstream &inf);
	int read_face_2_data( std::ifstream &inf);
	int read_version_A_face( std::ifstream &inf, sk3DRomFace &p_face);
	int read_version_B_face( std::ifstream &inf, sk3DRomFace &p_face);
	int read_vertex_data( std::istream &inf);
	int read_vertex_2_data( std::istream &inf);
	int read_vertex_3_data( std::istream &inf);
	int read_version_A_vertex_data( std::istream &inf);
	int read_texture_data( std::istream &inf);
	int read_skeleton_names( std::ifstream &inf, std::vector <char *> &sn);
	int read_skeleton( std::ifstream &inf, sk3DRomSkel &s);
	int read_main_skeleton_data( std::ifstream &inf);
	int read_skel_frames_2_data( std::ifstream &inf);
	int read_skel_animation_data( std::ifstream &inf);
	int read_animation_frame_data( std::ifstream &inf, const int sid);
	int read_model_texture_data( std::istream &inf);
	int read_texture_2_data( std::istream &inf);
	int read_texture_3_data( std::istream &inf);
	int read_bone_data( std::ifstream &inf);
	int read_bone_2_data( std::ifstream &inf);
	int read_basic_bone_data( std::ifstream &inf, sk3DRomBone &b);
	int read_bone_frame_data( std::ifstream &inf, sk3DRomBoneFrame &bf);
	int read_bone_frame_2_data( std::ifstream &inf, sk3DRomBoneFrame &bf);
	int read_group_data( std::ifstream &inf);
	int read_pages_data( std::ifstream &inf);
	int read_csg_data( std::ifstream &inf);

	//************ Saving sk3DRom Files ************//
	int write_vertex_data( std::ostream &des) const;
	int write_face_data( std::ostream &des) const;
	int write_main_skeleton_data( std::ostream &des) const;
	int write_skeleton_frame_data( std::ostream &des) const;
	int write_skeleton_animation_data( std::ostream &des) const;
	int write_skeleton_animation_frame_data( std::ostream &des,
		const int sid) const;
	int write_texture_data( std::ostream &des) const;
	int write_group_data( std::ostream &des) const;
	int write_pages_data( std::ostream &des) const;
	void write_page_faces( std::ostream &des,
		const sk3DRomTexturePage &p) const;
	int write_csg_data( std::ostream &des) const;

	void dbg_tv_check_leak();

	void hlp_render_faces_V(
		const bool use_group_color,
		const bool block_bsp_face, const int *flist,
		const int count,
		bool (*draw_test_fn)( const sk3DRomModel &, const int)) const;
	void hlp_render_faces_VN(
		const bool use_group_color,
		const bool block_bsp_face, const int *flist,
		const int count,
		bool (*draw_test_fn)( const sk3DRomModel &, const int)) const;
	void hlp_render_faces_VT(
		const bool use_group_color,
		const bool block_bsp_face, const int *flist,
		const int count,
		bool (*draw_test_fn)( const sk3DRomModel &, const int)) const;
	void hlp_render_faces_VNT(
		const bool use_group_color,
		const bool block_bsp_face, const int *flist,
		const int count,
		bool (*draw_test_fn)( const sk3DRomModel &, const int)) const;

private:
	sk3DRomModel( const sk3DRomModel &);
	sk3DRomModel &operator =( const sk3DRomModel &);
};

class sk3DRomVNTArray : public sk3DRomVertexArray
{
public:
	sk3DRomVNTArray();
	virtual ~sk3DRomVNTArray();

	int create_arrays( const sk3DRomModel &mdl);
	void update_arrays( const sk3DRomModel &mdl);
	void render_arrays();
	void delete_arrays();


protected:

private:
	int m_vertex_count;
	std::vector <sk3DRomVertexArray_RenderGroup> m_vertex_render_group;

	float *m_data;
	float *normal_offset( float *p_data) const { return p_data + 4;}
	float *fid_offset( float *p_data) const { return p_data + 7;}
	float *texcoord_offset( float *p_data) const { return p_data + 8;}
	unsigned short &fid_value( float *p_data) const {
		return *((unsigned short *)(fid_offset( p_data)));}

	void update_vertex( float *p, const sk3DRomModel &mdl,
		const sk3DRomFace &f, const int vid) const
	{
		mdl.fast_get_tv_vertex( f.vid[vid], p[0], p[1], p[2]);
		float *n = normal_offset( p);
		n[0] = f.n[vid].x;
		n[1] = f.n[vid].y;
		n[2] = f.n[vid].z;
	}
};

rgVector sk3DRom_get_bone_vector( sk3DRomBone &b);
rgVector sk3DRom_get_bone_vector( rgVector a, rgVector b, sk3DRomBone &bone);

void interpolate_bone_csys( const rgFloat w, const rgCoordSystem &src_cs,
	const rgCoordSystem &des_cs, rgCoordSystem &res_cs);

bool sk3DRomModel_face_draw_if_visible( const sk3DRomModel &mdl,
	const int fid);
bool sk3DRomModel_face_draw_if_mapped( const sk3DRomModel &mdl,
	const int fid);
bool sk3DRomModel_face_draw_if_unmapped( const sk3DRomModel &mdl,
	const int fid);

#endif
