/*
Copyright (C) 2004-2005 Amour Hassan

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

#include "rgProcess.h"
#include "rgParse.h"


inline bool char_is_in_string( const char c, char *s, int len);

rgPCommandHandler::rgPCommandHandler( const char *p_cname,
	int (*p_fn)( rgParseContext &))
{
	if( !p_cname) cname = 0;
	else {
		cname = new char[ strlen( p_cname) + 1];
		if( cname) strcpy( cname, p_cname);
	}

	fn = p_fn;
}

rgPCommandHandler::rgPCommandHandler( const rgPCommandHandler &c)
{
	copy_data( c);
}

rgPCommandHandler::~rgPCommandHandler()
{
	if( cname) delete [] cname;
}

rgPCommandHandler &rgPCommandHandler::operator =( rgPCommandHandler &c)
{
	copy_data( c);
	return *this;
}

void rgPCommandHandler::copy_data( const rgPCommandHandler &c)
{
	if( !c.cname) cname = 0;
	else {
		cname = new char[ strlen( c.cname) + 1];
		if( cname) strcpy( cname, c.cname);
	}

	fn = c.fn;
}



bool rgPNumeric::is_integer()
{
	switch( type) {
	case RGP_PNUMERIC_TYPE_INT:
	case RGP_PNUMERIC_TYPE_LONG:
		return true;
	}
	return false;
}

long rgPNumeric::get_integer()
{
	switch( type) {
	case RGP_PNUMERIC_TYPE_INT: return m_int;
	case RGP_PNUMERIC_TYPE_LONG: return m_long;
	}
	return 0;
}

double rgPNumeric::get_float()
{
	switch( type) {
	case RGP_PNUMERIC_TYPE_FLOAT: return m_float;
	case RGP_PNUMERIC_TYPE_DOUBLE: return m_double;
	}
	return 0.0;
}

rgParseContext::rgParseContext()
{
	init( RGP_DEFAULT_WORD_BUF_LEN);
}

rgParseContext::rgParseContext( int wblen)
{
	init( wblen);
}

void rgParseContext::init( int wblen)
{
	line = 1;
	error = 0;

	case_sensitive = false;
	append_word_terminators = false;

	/* The length of the word buffer must be at least 2. This will leave
	 * one character for one letter, and one character for the null
	 * terminator.
	 */
	if( wblen <= 2) wblen = 2;
	word = new char[ wblen];
	if( !word) {
		max_word_length = 0;
	} else {
		max_word_length = wblen;
	}

	data = 0;

	special_char_buf = new char[ 256 * 3];
	if( special_char_buf) {
		ws_chars = special_char_buf;
		wt_chars = special_char_buf + 256;
		ec_chars = special_char_buf + (256 * 2);

		ws_chars[0] = '\r';
		ws_chars[1] = '\t';
		ws_chars[2] = ' ';
		ws_chars[3] = '\0';
		ws_len = 4;

		wt_chars[0] = '\"';
		wt_chars[1] = '\0';
		wt_len = 2;

		ec_chars[0] = '\\';
		ec_len = 1;

	} else {
		ws_chars = wt_chars = ec_chars = 0;
		ws_len = wt_len = ec_len = 0;
	}

	parse_lock = false;
}

rgParseContext::~rgParseContext()
{
	clear();
	if( word) delete [] word;
	if( special_char_buf) delete [] special_char_buf;
}

void rgParseContext::clear()
{
	/* Close the file currently being parsed (if one exists). */
	if( inf.is_open()) inf.close();
	inf.clear();	/* Reset for reuse */

	line = 1;
	error = 0;

	case_sensitive = false;
	append_word_terminators = false;

	//if( word) { delete [] word; word = 0;}
	//data = 0;	/* Detach the extra parse data */

	/* Remove all command handlers. */
	while( !ch_fn.empty()) {
		if( (ch_fn[0])) delete (ch_fn[0]);
		ch_fn.erase( ch_fn.begin());
	}

	parse_lock = false;
}

bool rgParseContext::add_word_terminator_char( const char c)
{ return add_special_char( wt_chars, wt_len, c);}

bool rgParseContext::add_white_space_char( const char c)
{ return add_special_char( ws_chars, ws_len, c);}

bool rgParseContext::add_escape_char( const char c)
{ return add_special_char( ec_chars, ec_len, c);}

bool rgParseContext::add_special_char( char *&s, int &len, const char c)
{
	if( len < 256) { s[ len++] = c; return true;}
	return false;
}

int rgParseContext::add_command_handler( const char *cname,
	int (*fn)( rgParseContext &p))
{
	if( !cname) return -1;

	rgPCommandHandler *ch = new rgPCommandHandler( cname, fn);
	if( !ch) return -2;
	ch_fn.push_back( ch);
	
	return int( ch_fn.size() - 1);
}

char *rgParseContext::get_word_buffer()
{
	return word;
}

int rgParseContext::remove_command_handler( const int id)
{
	if( (id < 0) || (id >= int( ch_fn.size()))) return -1;

	if( (ch_fn[id])) delete (ch_fn[id]);
	ch_fn.erase( ch_fn.begin() + id);
	return 0;
}

// NOTE: Windows differs from Unix-Like operating systems when it comes to
//      the name of the "string compare ignoring case" function.
#ifdef WINDOWS_OS
#define STRCASECMP stricmp
#else
#define STRCASECMP strcasecmp
#endif
int rgParseContext::process_command( const char *cname)
{
	const int i_stop = int( ch_fn.size());
	if( case_sensitive) {
		/* Case sensitive command search. */
		for( int i = 0; i < i_stop; i++) {
			if( 0 == strcmp( cname, (ch_fn[i])->cname)) {
				if( (ch_fn[i])->fn) error =
					(ch_fn[i])->fn( *this);
				return error;
			}
		}
	} else {
		/* Case in-sensitive command search. */
		for( int i = 0; i < i_stop; i++) {
                        if( 0 == STRCASECMP( cname, (ch_fn[i])->cname)) {
				if( (ch_fn[i])->fn) error =
					(ch_fn[i])->fn( *this);
				return error;
			}
		}
	}

	return 0;
}
#undef STRCASECMP

int rgParseContext::get_word( char *des, const int len)
{
	return get_word( des, len, 0);
}

int rgParseContext::get_word( char *des, const int len, const int flags)
{
	if( !des) return -999;
	if( len <= 0) return 0;

	int i = 0;
	char c;
	bool word_char_read = false;
	while( 1) {
		c = inf.get();
		if( inf.eof()) {
			des[ i++] = '\0';
			if( 1 == i) return 0;
			return i;
		}
		if( inf.fail()) return -1;

		if( '\n' == c) {
			if( (flags & RGP_FLAG_NO_NEWLINE)) {
				inf.seekg( -1, std::ios::cur);
				return -3000;
			}

			if( (flags & RGP_FLAG_ONE_LINE)) {
				/* Put back the newline read since the caller
				 * requested to stay on the same line.
				 */
				inf.seekg( -1, std::ios::cur);
				des[ i++] = '\0';
				return i;
			}

			line++;
			if( word_char_read) {
				des[ i++] = '\0';
				return i;
			}

		} else if( char_is_in_string( c, ws_chars, ws_len)) {
			if( word_char_read) {
				des[ i++] = '\0';
				return i;
			}

		} else if( char_is_in_string( c, wt_chars, wt_len)) {
			terminating_char = c;
			if( append_word_terminators) des[ i++] = c;
			else {
				if( i >= 1) inf.seekg( -1, std::ios::cur);
				else des[ i++] = c;
			}
			des[ i++] = '\0';
			return i;

		} else if( char_is_in_string( c, ec_chars, ec_len)) {
			c = inf.get();
			if( inf.eof()) return -2;
			if( inf.fail()) return -3;
			des[ i++] = c;
			if( i == len) return -5;
			word_char_read = true;

		} else {
			des[ i++] = c;
			if( i == len) return -9;
			word_char_read = true;
		}
	}

	/* Execution never reaches this point. */
	return -12345;
}

int rgParseContext::parse( const char *fname)
{
	rgLocalSemaphor psem( &parse_lock);
	if( !psem.lock()) return -1;

	if( !fname) return -2;

	/* Initialize parsing variables/data. */
	line = 1;
	error = 0;

	/* Close the file currently being parsed (if one exists). */
	if( inf.is_open()) {
		inf.close();
		inf.clear();	/* Reset for reuse */
	}

	/* Open the new file to parse. */
	inf.open( fname, std::ios::binary);
	if( !inf.is_open()) return -3;

	int len;
	bool done = false;
	while( !done) {
		len = get_word( word, max_word_length);
		if( 0 == len) done = true;
		else if( len < 0) return len;

		error = process_command( word);
		if( error < 0) return error;
	}

	return 0;
}

int rgParseContext::skip_white_space()
{
	return skip_white_space( 0);
}

int rgParseContext::skip_white_space( const int flags)
{
	char c;
	while( 1) {
		c = inf.get();
		if( inf.eof()) return 0;
		if( inf.fail()) return -1;
		
		if( '\n' == c) {
			if( (flags & RGP_FLAG_NO_NEWLINE)) {
				inf.seekg( -1, std::ios::cur);
				return -3000;
			}
			if( (flags & RGP_FLAG_ONE_LINE)) {
				/* Newline '\n' does not count as white space
				 * in this case.
				 */
				inf.seekg( -1, std::ios::cur);
				return 0;
			}
			line++;

		} else if( !char_is_in_string( c, ws_chars, ws_len)) {
			inf.seekg( -1, std::ios::cur);
			return 0;
		}
	}

	/* Execution never reaches this point. */
	return -12345;
}

int rgParseContext::skip_line()
{
	char c;
	while( 1) {
		c = inf.get();
		if( inf.eof()) return 0;
		if( inf.fail()) return -1;

		if( '\n' == c) {
			line++;
			return 0;
		}
	}

	/* Execution never reaches this point. */
	return -12345;
}

int rgParseContext::set_string( char *des, const int len)
{
	if( !des) return -999;
	if( len <= 0) return -1000;

	int err = skip_white_space();
	if( err != 0) return -1;

	char c;
	c = inf.get();
	if( inf.eof() || inf.fail()) return -2;

	/* If expected quote was not found. */
	if( '"' != c) return -7;

	int i = 0;
	while( 1) {
		c = inf.get();

		if( '"' == c) {
			des[ i++] = '\0';
			return i;

		} else if( char_is_in_string( c, ec_chars, ec_len)) {
			c = inf.get();
			if( inf.eof()) return -10;
			if( inf.fail()) return -11;
			des[ i++] = c;
			if( len == i) return -13;
		
		} else {
			des[ i++] = c;
			if( len == i) return -15;
		
		}
	}

	/* Execution never reaches this point. */
	return -12345;
}

int rgParseContext::set_numeric( rgPNumeric &des)
{
	return set_numeric( des, 0);
}

inline void hlp_seek_back_unused( rgParseContext &p, char *s, char *ep,
	const int seek_start)
{
	const int k = (ep - s);
	if( !k) return;
	p.inf.seekg( seek_start + k, std::ios::beg);
}

int rgParseContext::set_numeric( rgPNumeric &des, const int flags)
{
	des.type = RGP_PNUMERIC_TYPE_INVALID;

	int err = skip_white_space( flags);
	if( err != 0) return -1;

	const int seek_start = inf.tellg();
	int len = get_word( word, max_word_length, flags);
	/* REMEMBER: "len" includes the null terminator, thus even an empty
	 *	string will be length one (otherwise there is a serios error).
	 */
	if( len <= 1) return -5;

	/* Determine the base of the number read. */
	bool result;
	bool num_read = false;
	char *ep;	/* "EndPtr", tells how much of string was read. */
	const bool has_decimal_point = str_numeric_has_decimal_point( word);

	if( len >= 2) {
		/* If the number begins with a zero and does not have a
		 * decimal point in it, then it will be treated as being an
		 * octal number, or it will be a hexidecimal number if the
		 * zero is followed by 'x' or 'X'.
		 */
		if( ('0' == word[0]) && !has_decimal_point) {

			num_read = true;
			/* If the number is in hexadecimal format. */
			switch( word[1]) {
			case 'X':
			case 'x':
				result = str_to_int( des.m_int, word, 16, ep);
				if( !result) return -25;
				break;

			default:
				result = str_to_int( des.m_int, word, 8, ep);
				if( !result) return -26;
				break;
			}
			if( num_read) des.type = RGP_PNUMERIC_TYPE_INT;
			hlp_seek_back_unused( *this, word, ep, seek_start);
			return 0;
		}
	}

	if( !num_read) {
		/* The number is in decimal (base 10) format. */
		if( has_decimal_point) {
			switch( word[ len - 1]) {
			case 'F':
			case 'f':
				result = str_to_float( des.m_float, word, ep);
				if( !result) return -35;
				des.type = RGP_PNUMERIC_TYPE_FLOAT;
				break;

			default:
				result = str_to_double( des.m_double, word,
					ep);
				if( !result) return -37;
				des.type = RGP_PNUMERIC_TYPE_DOUBLE;
				break;
			}
			hlp_seek_back_unused( *this, word, ep, seek_start);
			return 0;
		} else {
			result = str_to_int( des.m_int, word, 10, ep);
			if( !result) return -40;
			des.type = RGP_PNUMERIC_TYPE_INT;
			hlp_seek_back_unused( *this, word, ep, seek_start);
			return 0;
		}
	}
	
	return 0;
}

bool rgParseContext::str_numeric_has_decimal_point( char *num)
{
	while( (*num)) { if( '.' == (*(num++))) return true;}
	return false;
}

bool rgParseContext::str_to_int( int &des, char *src, int base, char *&endptr)
{
	des = strtol( src, &endptr, base);
	if( endptr != src) return true;
	return false;
}

bool rgParseContext::str_to_long( long &des, char *src, int base,
	char * &endptr)
{
	des = strtol( src, &endptr, base);
	if( endptr != src) return true;
	return false;
}

bool rgParseContext::str_to_float( float &des, char *src, char *&endptr)
{
	/* FIXME: "strtof" is not defined in "stdlib.h"!!! */
	// des = strtof( src, &endptr);
	des = strtod( src, &endptr);
	if( endptr != src) return true;
	return false;
}

bool rgParseContext::str_to_double( double &des, char *src, char *&endptr)
{
	des = strtod( src, &endptr);
	if( endptr != src) return true;
	return false;
}

int rgParseContext::read_expected_char( const char c, bool one_line)
{
	int error;
	if( one_line) error = skip_white_space( RGP_FLAG_NO_NEWLINE);
	else error = skip_white_space();
	if( error) return -1;

	const char temp = inf.get();
	if( inf.fail()) return -2;
	if( c != temp) return -3;
	return 0;
}

inline bool char_is_in_string( const char c, char *s, int len)
{
	while( len--) if( c == (*(s++))) return true;
	return false;
}





